import os
import psutil
import re
import sys
import yaml
from os import listdir
from os.path import isfile, join, abspath, splitext, basename

import falcon
from falcon_auth import FalconAuthMiddleware, TokenAuthBackend
from paste.translogger import TransLogger
from waitress import serve

from .datastore import MetricsDataStoreManager
from prometheus_client import CONTENT_TYPE_LATEST


CONFIG_DEFAULTS = {
    'metrics_path': '/var/lib/metrics',
    'auth_token': None,
    'host': '0.0.0.0',
    'port': 8902,
    'debug': False
}


def load_config(config_file=None, config_args={}):
    config = {}
    if os.path.exists(config_file):
        with open(config_file, encoding='utf-8') as f:
            config = yaml.safe_load(f)

    return {**CONFIG_DEFAULTS, **config, **config_args}


class ExporterResource:
    def __init__(self, config):
        self.config = config

    def on_get(self, req, resp):
        data = {}
        filters = {}
        for param, value in req.params.items():
            if param.startswith('filter_'):
                p = param.split("filter_")[1]
                filters[p] = value

        key = req.params.get('key')
        key_type = req.params.get('key_type') or None

        for mf in self._get_metric_files(key_type):
            filename = basename(mf)
            name = splitext(filename)[0]
            try:
                manager = MetricsDataStoreManager.deserialize(
                    mf, filters=filters)
                data[name] = manager.dump_prometheus_metrics()
            except Exception as e:
                continue

        # if we have a key filter only that 'type'
        # else convert everything to a response
        res = b''.join(map(bytes, data.values()))

        if key:
            if key in data:
                res = data[key]
            else:
                res = b''

        if self.config['debug']:
            process = psutil.Process(os.getpid())
            print("Memory usage %sMB" %
                  round(process.memory_info().rss / 1000000, 2), flush=True)
        resp.content_type = CONTENT_TYPE_LATEST
        resp.data = res

    def _get_metric_files(self, key_type=None):
        metrics_path = self.config['metrics_path']
        if key_type:
            metrics_path = os.path.join(self.config['metrics_path'], key_type)
        metric_files = [
            abspath(join(metrics_path, f))
            for f in listdir(metrics_path)
            if isfile(join(metrics_path, f))]
        return metric_files


class Exporter:
    def __init__(self, config_file=None, config_args={}):
        auth_backend = TokenAuthBackend(
            self._user_loader, auth_header_prefix='Bearer')
        auth_middleware = FalconAuthMiddleware(auth_backend)
        self.app = falcon.App(middleware=[auth_middleware])

        self.config = load_config(
            config_file=config_file, config_args=config_args)
        exporter = ExporterResource(self.config)

        # Routing
        self.app.add_route('/', exporter)
        # Adding a request logger as wsgi middleware
        self.app = TransLogger(self.app)

    def _user_loader(self, token):
        if not self.config['auth_token']:
            return {'username': 'no-auth'}
        if self.config['auth_token'] == token:
            return {'username': 'prometheus'}
        return None

    def run(self):
        print("Starting server on %s:%s" %
              (self.config['host'], self.config['port']))
        serve(self.app, host=self.config['host'], port=self.config['port'])
